"""
Data simulation Quantiphyse plugin

Generic classes for data and structural models

Author: Martin Craig <martin.craig@eng.ox.ac.uk>
Copyright (c) 2016-2017 University of Oxford, Martin Craig
"""

from __future__ import division, unicode_literals, absolute_import, print_function

from quantiphyse.data.qpdata import Metadata
from quantiphyse.utils import QpException, LogSource

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

class Parameter:
    def __init__(self, name, display_name=None, **kwargs):
        self.name = name
        if display_name is None:
            display_name = name
        self.display_name = display_name
        self.kwargs = kwargs

class Model(LogSource):
    """
    """
    def __init__(self, ivm, display_name, **kwargs):
        LogSource.__init__(self)
        self._ivm = ivm
        self.display_name = display_name
        self.options = Metadata()



"""
Data simulation Quantiphyse plugin

Data simulation process

Author: Martin Craig <martin.craig@eng.ox.ac.uk>
Copyright (c) 2016-2017 University of Oxford, Martin Craig
"""

import numpy as np

from quantiphyse.data import NumpyData
from quantiphyse.processes import Process
from quantiphyse.utils import QpException

from .data_models import get_data_models
from .struc_models import get_struc_models

class DataSimProcess(Process):
    """
    Calculate CBF from R2P / DBV output
    """
    
    PROCESS_NAME = "DataSim"
    
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)
        self._struc_models = get_struc_models()
        self._data_models = get_data_models()

    def run(self, options):
        data_model_name = options.pop("data-model", None)
        if data_model_name is None:
            raise QpException("Data model not specified")
        data_model = self._data_models.get(data_model_name, None)
        if data_model is None:
            raise QpException("Unknown data model: %s" % data_model_name)
        data_model = data_model(self.ivm)
        data_model_options = options.pop("data-model-options", {})
        data_model.options = data_model_options
        self.log("Created data model: %s\n" % data_model_name)

        struc_model_name = options.pop("struc-model", None)
        struc_model_options = options.pop("struc-model-options", {})
        if struc_model_name is None:
            raise QpException("Structure model not specified")
        struc_model = self._struc_models.get(struc_model_name, None)
        if struc_model is None:
            raise QpException("Unknown structure model: %s" % struc_model_name)
        struc_model = struc_model(self.ivm)
        struc_model.options = struc_model_options
        self.log("Created structure model: %s\n" % struc_model_name)

        param_values = options.pop("param-values", {})
        output_param_maps = options.pop("output-param-maps", False)
        self.log("Getting simulated data\n")
        ret = struc_model.get_simulated_data(data_model, param_values, output_param_maps=output_param_maps)
        if output_param_maps:
            sim_data, param_maps = ret
        else:
            sim_data, param_maps = ret, {}

        output_clean_name = options.pop("output-clean", "")
        if output_clean_name:
            self.ivm.add(sim_data.raw().copy(), grid=sim_data.grid, name=output_clean_name, make_current=False)

        for param, qpdata in param_maps.items():
            self.ivm.add(qpdata, name=param, make_current=False)

        output_name = options.pop("output", "sim_data")
        self.ivm.add(sim_data, name=output_name, make_current=True)

        output_roi = output_name + "_roi"
        if sim_data.ndim > 3:
            roi_data = sim_data.raw()[..., 0]
        else:
            roi_data = sim_data.raw()
        roi = NumpyData(np.array(roi_data > 0, dtype=np.int), grid=sim_data.grid, roi=True, name=output_roi)
        self.ivm.add(roi, make_current=False)



"""
Data simulation Quantiphyse plugin

Author: Martin Craig <martin.craig@eng.ox.ac.uk>
Copyright (c) 2016-2017 University of Oxford, Martin Craig
"""

from __future__ import division, unicode_literals, absolute_import, print_function

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import QpWidget, Citation, TitleWidget, RunWidget, WarningBox
from quantiphyse.gui.options import OptionBox, DataOption, NumericOption, BoolOption, NumberListOption, TextOption, ChoiceOption
from quantiphyse.utils import QpException

from ._version import __version__
from .data_models import get_data_models
from .struc_models import get_struc_models
from . import data_model_views
from . import struc_model_views

def get_view_class(model_class):
    view_name = model_class.__name__ + "View"
    cls = getattr(data_model_views, view_name, None)
    if cls is None:
        cls = getattr(struc_model_views, view_name, None)
    return cls

class ParamValuesGrid(QtGui.QGroupBox):
    """
    Widget which presents a grid of values so the user can specify the value of 
    each parameter in each partial volume structure
    """

    def __init__(self, title="Parameter Values"):
        QtGui.QGroupBox.__init__(self, title)
        self._grid = QtGui.QGridLayout()
        self.setLayout(self._grid)
        self._structures = []
        self._params = []
        self._values = {}
        
    @property
    def structures(self):
        """
        A sequence of named structures
        """
        return self._structures

    @structures.setter
    def structures(self, structures):
        structures = sorted(structures, key=lambda x: x.name)
        if structures != self._structures:
            self._structures = structures
            self._repopulate()

    @property
    def params(self):
        """
        A sequence of named parameters
        """
        return self._params

    @params.setter
    def params(self, params):
        if params != self._params:
            self._params = params
            self._repopulate()

    @property
    def values(self):
        """
        Mapping from structure name to another mapping from param name
        to parameter value(s) in that structure
        """
        ret = {}
        for structure_idx, structure in enumerate(self._structures):
            ret[structure.name] = {}
            if structure.name not in self._values:
                self._values[structure.name] = {}
            for param_idx, param in enumerate(self._params):
                item = self._grid.itemAtPosition(param_idx + 1, structure_idx + 1)
                vals = item.widget().value
                if len(vals) == 1:
                    vals = vals[0]
                self._values[structure.name][param.name] = vals

                # We keep records in self._values for data structures/params that aren't in the
                # current list, so need to be careful to only return the ones that are
                ret[structure.name][param.name] = vals

        return ret

    def _repopulate(self):
        self._clear()
        for structure_idx, structure in enumerate(self._structures):
            self._grid.addWidget(QtGui.QLabel(structure.display_name), 0, structure_idx + 1)

        for param_idx, param in enumerate(self._params):
            self._grid.addWidget(QtGui.QLabel(param.display_name), param_idx + 1, 0)
            for structure_idx, structure in enumerate(self._structures):
                # The initial value for a parameter in a structure is the previous value set (if any), 
                # the default value for that structure (if any), the default value for the parent
                # structure (if any), the generic parameter default (if any), and if none of the above, zero.
                if structure.name in self._values and param.name in self._values[structure.name]:
                    initial = self._values[structure.name][param.name]
                elif structure.name in param.kwargs.get("struc_defaults", {}):
                    initial = param.kwargs["struc_defaults"][structure.name]
                elif structure.kwargs.get("parent_struc", "") in param.kwargs.get("struc_defaults", {}):
                    initial = param.kwargs["struc_defaults"][structure.kwargs.get("parent_struc", "")]
                else:
                    initial = [param.kwargs.get("default", 0.0)]
                if isinstance(initial, (int, float)):
                    initial = [initial]
                self._grid.addWidget(NumberListOption(initial=initial,
                                                      intonly=param.kwargs.get("intonly", False),
                                                      load_btn=False), param_idx + 1, structure_idx + 1)

    def _clear(self):
        while self._grid.count():
            child = self._grid.takeAt(0)
            if child.widget():
                child.widget().deleteLater()

class OptionsWidget(QtGui.QWidget):

    sig_changed = QtCore.Signal()

    def __init__(self, ivm, parent):
        QtGui.QWidget.__init__(self, parent)
        self.ivm = ivm

class ModelOptions(OptionsWidget):
    def __init__(self, ivm, parent, model_type, abbrev, model_classes):
        OptionsWidget.__init__(self, ivm, parent)
        self._views = {}
        self.model = None
        self.view = None
        self._option_name = "%s-model" % abbrev
        for name, cls in model_classes.items():
            self._views[name] = get_view_class(cls)(ivm)

        main_vbox = QtGui.QVBoxLayout()
        self.setLayout(main_vbox)

        self.options = OptionBox()
        self.options.add("%s model" % model_type, ChoiceOption([v.model.display_name for v in self._views.values()], self._views.keys()), key=self._option_name)
        self.options.option(self._option_name).sig_changed.connect(self._model_changed)
        main_vbox.addWidget(self.options)

        self._create_guis(main_vbox)
        main_vbox.addStretch(1)
        self._model_changed()

    def _create_guis(self, main_vbox):
        # Create the GUIs for models - only one visible at a time!
        for view in self._views.values():
            if view.gui is not None:
                view.gui.setVisible(False)
                if isinstance(view.gui, QtGui.QWidget):
                    main_vbox.addWidget(view.gui)
                else:
                    main_vbox.addLayout(view.gui)
                view.model.options.sig_changed.connect(self._model_option_changed)
                
    def _model_changed(self):
        chosen_name = self.options.option(self._option_name).value
        self.view = self._views[chosen_name]
        self.model = self.view.model
        for name, view in self._views.items():
            view.gui.setVisible(chosen_name == name)
        self.sig_changed.emit()

    def _model_option_changed(self, _key, _value):
        self.sig_changed.emit()

class NoiseOptions(OptionsWidget):
    def __init__(self, ivm, parent):
        OptionsWidget.__init__(self, ivm, parent)

        main_vbox = QtGui.QVBoxLayout()
        self.setLayout(main_vbox)

        self.options = OptionBox()
        self.options.add("Add noise with SNR", NumericOption(minval=0.1, maxval=100, default=10), checked=True, key="snr")
        self.options.sig_changed.connect(self.sig_changed.emit)
        main_vbox.addWidget(self.options)

        main_vbox.addStretch(1)

class MotionOptions(OptionsWidget):
    def __init__(self, ivm, parent):
        OptionsWidget.__init__(self, ivm, parent)

        main_vbox = QtGui.QVBoxLayout()
        self.setLayout(main_vbox)

        self.options = OptionBox()
        self.options.add("Simulate motion", BoolOption(default=False), key="motion")
        self.options.option("motion").sig_changed.connect(self._update_widget_visibility)
        self.options.add("Random translation standard deviation (mm)", NumericOption(minval=0, maxval=5, default=1, decimals=2), key="std")
        self.options.add("Random rotation standard deviation (\N{DEGREE SIGN})", NumericOption(minval=0, maxval=10, default=1, decimals=2), key="std_rot")
        self.options.add("Padding (mm)", NumericOption(minval=0, maxval=10, default=5, decimals=1), key="padding", checked=True)
        self.options.add("Interpolation", ChoiceOption(["Nearest neighbour", "Linear", "Quadratic", "Cubic"], return_values=range(4), default=3), key="order")
        main_vbox.addWidget(self.options)

        main_vbox.addStretch(1)
        self._update_widget_visibility()

    def _update_widget_visibility(self):
        enabled = self.options.option("motion").value
        for option in ["std", "std_rot", "padding", "order"]:
            self.options.set_visible(option, enabled)

class ParamsOptions(OptionsWidget):
    def __init__(self, ivm, parent):
        OptionsWidget.__init__(self, ivm, parent)

        main_vbox = QtGui.QVBoxLayout()
        self.setLayout(main_vbox)

        self._params = ParamValuesGrid()
        main_vbox.addWidget(self._params)
        main_vbox.addStretch(1)

    def set_models(self, data_model, struc_model):
        self._params.params = data_model.params
        self._params.structures = struc_model.structures

    @property
    def options(self):
        return self._params.values

class OutputOptions(OptionsWidget):
    def __init__(self, ivm, parent):
        OptionsWidget.__init__(self, ivm, parent)

        main_vbox = QtGui.QVBoxLayout()
        self.setLayout(main_vbox)

        self.options = OptionBox()
        self.options.add("Output name", TextOption("sim_data"), key="output")
        self.options.add("Output parameter maps", BoolOption(), default=False, key="output-param-maps")
        self.options.add("Output clean data (no noise/motion)", TextOption("sim_data_clean"), checked=True, default=True, key="output-clean")
        main_vbox.addWidget(self.options)
        main_vbox.addStretch(1)

class DataSimWidget(QpWidget):
    """
    Data simulation widget
    """
    def __init__(self, **kwargs):
        QpWidget.__init__(self, name="Data Simulator", icon="datasim", group="Simulation",
                          desc="Simulates data for various imaging sequences, from known parameter inputs", **kwargs)
        self._param_values = {}

    def init_ui(self):
        main_vbox = QtGui.QVBoxLayout()
        self.setLayout(main_vbox)

        title = TitleWidget(self, help="generic/datasim", subtitle="Simulates data for various imaging sequences, from known parameter inputs %s" % __version__)
        main_vbox.addWidget(title)

        self.tabs = QtGui.QTabWidget()
        main_vbox.addWidget(self.tabs)

        self.struc_model = ModelOptions(self.ivm, parent=self, model_type="Structural", abbrev="struc", model_classes=get_struc_models())
        self.struc_model.sig_changed.connect(self._update_params)
        self.data_model = ModelOptions(self.ivm, parent=self, model_type="Data", abbrev="data", model_classes=get_data_models())
        self.data_model.sig_changed.connect(self._update_params)
        self.params = ParamsOptions(self.ivm, parent=self)
        self.noise = NoiseOptions(self.ivm, parent=self)
        self.motion = MotionOptions(self.ivm, parent=self)
        self.output = OutputOptions(self.ivm, parent=self)
        self.tabs.addTab(self.struc_model, "Structure model")
        self.tabs.addTab(self.data_model, "Data model")
        self.tabs.addTab(self.params, "Parameter values")
        self.tabs.addTab(self.noise, "Noise")
        self.tabs.addTab(self.motion, "Motion")
        self.tabs.addTab(self.output, "Output")

        main_vbox.addStretch(1)
        main_vbox.addWidget(RunWidget(self))
        self._update_params()

    def _update_params(self):
        self.params.set_models(self.data_model.model, self.struc_model.model)

    def processes(self):
        processes = []
        opts = self.output.options.values()
        opts.update(self.struc_model.options.values())
        opts.update(self.data_model.options.values())
        opts["struc-model-options"] = self.struc_model.model.options
        opts["data-model-options"] = self.data_model.model.options
        opts["param-values"] = self.params.options
        processes.append({"DataSim" : opts})

        motion_opts = self.motion.options.values()
        if motion_opts.pop("motion"):
            motion_opts.update({
                "data" : opts["output"],
                "output-name" : opts["output"],
            })
            processes.append({"SimMotion" : motion_opts})

        noise_opts = self.noise.options.values()
        if "snr" in noise_opts:
            noise_opts.update({
                "data" : opts["output"],
                "roi" : opts["output"] + "_roi",
                "output-name" : opts["output"],
            })
            # Hack to support ASL TC data where the 'signal' is the difference
            # not the magnitude of the static signal
            if opts["data-model"] == "asl" and opts["data-model-options"].get("iaf", "diff") in ("tc", "ct"):
                noise_opts["mode"] = "diff"
            processes.append({"AddNoise" : noise_opts})

        processes.append({"Delete" : {opts["output"] + "_roi" : ""}})

        return processes


"""
Data simulation Quantiphyse plugin

GUIs for structural models, i.e. QtWidget instances which control the options
of a structural model

Author: Martin Craig <martin.craig@eng.ox.ac.uk>
Copyright (c) 2016-2017 University of Oxford, Martin Craig
"""

from __future__ import division, unicode_literals, absolute_import, print_function

import time

import numpy as np

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data import NumpyData, DataGrid, ImageVolumeManagement
from quantiphyse.gui.options import OptionBox, DataOption, NumericOption, BoolOption, NumberListOption, TextOption, ChoiceOption, RunButton
from quantiphyse.utils import QpException, get_plugins
from quantiphyse.processes import Process

from .struc_models import *

class AddEmbeddingDialog(QtGui.QDialog):
    """
    Dialog box enabling one item to be chosen from a list
    """

    def __init__(self, parent, ivm, existing_strucs):
        super(AddEmbeddingDialog, self).__init__(parent)
        self.ivm = ivm
        self.sel_text = None
        self.sel_data = None
        self.existing_names = [struc.name for struc in existing_strucs]

        self.setWindowTitle("Add embedding")
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        self._opts = OptionBox()
        pvmap = self._opts.add("PV map / mask", DataOption(ivm, data=True, rois=True), key="pvmap")
        pvmap.sig_changed.connect(self._pvmap_changed)
        self._opts.add("ROI region", ChoiceOption([]), key="region")
        name = self._opts.add("Name of embedded structure", TextOption(), key="name")
        name.textChanged.connect(self._name_changed)
        self._opts.add("Structure type", ChoiceOption(["Embedding", "Activation mask", "Additional PVE"], return_values=["embed", "act", "add"]), key="type")
        self._opts.add("Parent structure", ChoiceOption([s.display_name for s in existing_strucs], [s.name for s in existing_strucs]), key="parent")
        self._opts.add("Edge smoothing sigma (mm)", NumericOption(minval=0, maxval=10, default=1, decimals=2), checked=True, enabled=False, key="sigma")
        vbox.addWidget(self._opts)

        self._warning = QtWidgets.QLabel(parent=self)
        self._warning.setStyleSheet("color: red;")
        vbox.addWidget(self._warning)

        self.button_box = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        self.button_box.button(QtGui.QDialogButtonBox.Ok).setEnabled(False)
        vbox.addWidget(self.button_box)
        self._pvmap_changed()

    def _pvmap_changed(self):
        if self.pvmap:
            qpdata = self.ivm.data[self.pvmap]
            if not self.name:
                self._opts.option("name").value = qpdata.name
            if qpdata.roi:
                regions = list(qpdata.regions.keys())
                region_names = list(qpdata.regions.values())
                self._opts.option("region").setChoices(region_names, regions)
            self._opts.set_visible("region", qpdata.roi and len(qpdata.regions) > 1)
            self._opts.set_visible("sigma", qpdata.roi)

    def _name_changed(self):
        accept = self.name != "" and self.name not in self.existing_names
        self.button_box.button(QtGui.QDialogButtonBox.Ok).setEnabled(accept)
        if self.name and not accept:
            self._warning.setText("Name already exists")
        else:
            self._warning.setText("")

    @property
    def pvmap(self):
        return self._opts.option("pvmap").value

    @property
    def region(self):
        return self._opts.values().get("region", None)

    @property
    def sigma(self):
        return self._opts.values().get("sigma", None)

    @property
    def name(self):
        return self._opts.option("name").value

    @property
    def struc_type(self):
        return self._opts.option("type").value

    @property
    def parent_struc(self):
        return self._opts.option("parent").value

class UserPvModelView:
    """
    View for UserPvModel - a structural model where user supplies partial volume maps
    """

    def __init__(self, ivm):
        self.ivm = ivm
        self.model = UserPvModel(ivm)
        self.model.options.update({
            "pvmaps" : {},
            "additional" : {},
        })
        self.gui = OptionBox()
        self.gui.sig_changed.connect(self._update_options)
        self._refresh_gui()

    def _refresh_gui(self):
        options = dict(self.model.options)
        self.gui.clear()
        for struc in self.model.default_strucs:
            data_opt = self.gui.add("%s map" % struc.name.upper(), DataOption(self.model._ivm, explicit=True), checked=True, enabled=struc.name in options["pvmaps"], key=struc.name)
            data_opt.value = options["pvmaps"].get(struc.name, None)

        for struc in options["additional"].values():
            del_btn = self._struc_delete_btn(struc)
            display_type = {"add" : "map", "embed" : "embedding", "act" : "mask"}.get(struc["struc_type"], "map")
            data_opt = self.gui.add("%s %s" % (struc["name"], display_type), DataOption(self.model._ivm, explicit=True, rois=True), del_btn, key=struc["name"])
            data_opt.value = struc.get("pvmap", None)

        res_opts = options.get("resampling", {})
        self.gui.add("Resampling", ChoiceOption(["Downsample", "From another data set", "Specified resolution"], ["down", "data", "res"], 
                     default=res_opts.get("type", "res")), checked=True, enabled="type" in res_opts, key="type")
        self.gui.add("Output space from", DataOption(self.ivm), key="grid")
        self.gui.add("Output resample factor", NumericOption(intonly=True, minval=1, maxval=10, default=2), key="factor")
        self.gui.add("Voxel sizes (mm)", NumberListOption(load_btn=False), key="voxel-sizes")
        for opt in ("grid", "factor", "voxel-sizes"):
            if opt in res_opts:
                self.gui.option(opt).value = res_opts[opt]

        self.gui.add(None, RunButton("Add user-defined structure", callback=self._add_embedding), key="add_embedding")
        self._update_resamp_visibility()

    def _update_resamp_visibility(self):
        resample_type = self.gui.values().pop("type", "")
        self.gui.set_visible("grid", resample_type == "data")
        self.gui.set_visible("factor", resample_type in ("up", "down"))
        #self.gui.set_visible("order", resample_type != "down")
        #self.gui.set_visible("2d", resample_type in ("up", "down"))
        self.gui.set_visible("voxel-sizes", resample_type == "res")

    def _update_options(self):
        self._update_resamp_visibility()
        opts = self.gui.values()
        pvmaps = {}
        for struc in self.model.default_strucs:
            if struc.name in opts:
                pvmaps[struc.name] = opts[struc.name]
        self.model.options["pvmaps"] = pvmaps

        resamp = {}
        for resamp_opt in ("type", "grid", "factor", "voxel-sizes"):
            if resamp_opt in opts:
                resamp[resamp_opt] = opts[resamp_opt]
        self.model.options["resampling"] = resamp

    def _struc_delete_btn(self, add_struc):
        def _del_cb():
            self._del_struc(add_struc["name"])

        btn = QtGui.QPushButton("Delete")
        btn.clicked.connect(_del_cb)
        return btn

    def _del_struc(self, name):
        self.model.options["additional"].pop(name, None)
        self._refresh_gui()

    def _add_embedding(self):
        dialog = AddEmbeddingDialog(self.gui, self.model._ivm, self.model.default_strucs)
        try:
            accept = dialog.exec_()
        except:
            import traceback
            traceback.print_exc()
        if accept:
            self.model.options["additional"][dialog.name] = {
                "name" : dialog.name,
                "struc_type" : dialog.struc_type,
                "parent_struc" : dialog.parent_struc,
                "pvmap" : dialog.pvmap,
            }
            if dialog.region:
                self.model.options["additional"][dialog.name]["region"] = dialog.region
            if dialog.sigma:
                self.model.options["additional"][dialog.name]["sigma"] = dialog.sigma

            self._refresh_gui()

class FastStructureModelView:
    """
    View for FastStructureModel
    """
    def __init__(self, ivm):
        self.model = FastStructureModel(ivm)
        self.gui = OptionBox()
        self.gui.add("Structural image (brain extracted)", DataOption(self.model._ivm, explicit=True), key="struc")
        self.gui.add("Image type", ChoiceOption(["T1 weighted", "T2 weighted", "Proton Density"], return_values=[1, 2, 3]), key="type")
        self.gui.sig_changed.connect(self._update_options)
        
    def _update_options(self):
        self.model.options.update(self.gui.values())

class CheckerboardModelView:
    """
    View for CheckerboardModel
    """
    def __init__(self, ivm):
        self.model = CheckerboardModel(ivm)
        self.gui = OptionBox()
        self.gui.add("Number of voxels per patch (approx)", NumericOption(minval=1, maxval=1000, default=20, intonly=True), key="voxels-per-patch")
        self.gui.sig_changed.connect(self._update_options)
        
    def _update_options(self):
        self.model.options.update(self.gui.values())




"""
Datausion simulation Quantiphyse plugin

Data models, i.e. classes which generated simulated data for different
types of imaging

Author: Martin Craig <martin.craig@eng.ox.ac.uk>
Copyright (c) 2016-2017 University of Oxford, Martin Craig
"""

from __future__ import division, unicode_literals, absolute_import, print_function

import logging

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

import numpy as np

from fabber import Fabber

from quantiphyse.data import NumpyData
from quantiphyse.gui.options import OptionBox, DataOption, NumericOption, BoolOption, NumberListOption, TextOption, ChoiceOption
from quantiphyse.utils import get_plugins, QpException

from .data_models import *

LOG = logging.getLogger(__name__)

class SpinEchoDataModelView:
    def __init__(self, ivm):
        self.model = SpinEchoDataModel(ivm)
        self.gui = OptionBox()
        self.gui.add("TR (s)", NumericOption(minval=0, maxval=10, default=4.8), key="tr")
        self.gui.add("TE (ms)", NumericOption(minval=0, maxval=1000, default=0), key="te")
        self.gui.add("M0", NumericOption(minval=0, maxval=10000, default=1000), key="m0")
        self.gui.sig_changed.connect(self._update_options)
        self._update_options()

    def _update_options(self):
        self.model.options.update(self.gui.values())

class AslDataModelView:
    def __init__(self, ivm):
        self.model = AslDataModel(ivm)
        self.gui = OptionBox()
        self.gui.add("Bolus duration", NumericOption(minval=0, maxval=5, default=1.8), key="tau")
        self.gui.add("Labelling", ChoiceOption(["CASL/pCASL", "PASL"], [True, False], default=True), key="casl")
        self.gui.add("PLDs", NumberListOption([0.25, 0.5, 0.75, 1.0, 1.25, 1.5]), key="plds")
        self.gui.add("Time per slice (ms)", NumericOption(minval=0, maxval=1000, default=0, intonly=True), key="slicedt")
        self.gui.add("Data format", ChoiceOption(["Differenced data", "Label/Control pairs"], ["diff", "tc"]), key="iaf")
        self.gui.add("Repeats", NumericOption(minval=1, maxval=100, default=1, intonly=True), key="repeats")
        self.gui.add("Group by", ChoiceOption(["PLDs", "Repeats"], ["tis", "rpt"]), key="ibf")
        self.gui.add("Inversion efficiency", NumericOption(minval=0.5, maxval=1.0, default=0.85), key="alpha")
        self.gui.add("M0", NumericOption(minval=0, maxval=2000, default=1000), key="m0")
        self.gui.add("TR (s)", NumericOption(minval=0, maxval=10, default=4), key="tr")
        self.gui.add("TE (ms)", NumericOption(minval=0, maxval=1000, default=13), key="te")
        self.gui.add("Tissue/arterial partition coefficient", NumericOption(minval=0, maxval=1, default=0.9), key="pct")
        #self.gui.add("Arterial component", BoolOption(), key="incart")
        self.gui.sig_changed.connect(self._update_options)
        self._update_options()

    def _update_options(self):
        self.model.options.update(self.gui.values())

class DscDataModelView:
    def __init__(self, ivm):
        self.model = DscDataModel(ivm)
        self.gui = OptionBox()
        self.gui.add("Time between volumes (s)", NumericOption(minval=0, maxval=5, default=1.0), key="delt")
        self.gui.add("TE (s)", NumericOption(minval=0, maxval=5, default=1.0), key="te")
        self.gui.add("AIF", NumberListOption(), key="aif")
        self.gui.sig_changed.connect(self._update_options)
        self._update_options()

    def _update_options(self):
        self.model.options.update(self.gui.values())
         
class DceDataModelView:
    def __init__(self, ivm):
        self.model = DceDataModel(ivm)
        self.gui = OptionBox()
        self.gui.add("Model", ChoiceOption(["Standard Tofts model",
                                             "Extended Tofts model (ETM)",
                                             "2 Compartment exchange model",
                                             "Compartmental Tissue Update (CTU) model",
                                             "Adiabatic Approximation to Tissue Homogeneity (AATH) Model"],
                                            ["dce_tofts",
                                             "dce_ETM",
                                             "dce_2CXM",
                                             "dce_CTU",
                                             "dce_AATH"]), key="model")
        self.gui.add("Contrast agent R1 relaxivity (l/mmol s)", NumericOption(minval=0, maxval=10, default=3.7), key="r1")
        self.gui.add("Flip angle (\N{DEGREE SIGN})", NumericOption(minval=0, maxval=90, default=12), key="fa")
        self.gui.add("TR (ms)", NumericOption(minval=0, maxval=10, default=4.108), key="tr")
        self.gui.add("Time between volumes (s)", NumericOption(minval=0, maxval=30, default=12), key="delt")
        self.gui.add("AIF", ChoiceOption(["Population (Orton 2008)", "Population (Parker)", "Measured DCE signal", "Measured concentration curve"], ["orton", "parker", "signal", "conc"]), key="aif")
        self.gui.add("Number of volumes", NumericOption(minval=0, maxval=100, default=20, intonly=True), key="nt")
        self.gui.add("Bolus injection time (s)", NumericOption(minval=0, maxval=60, default=30), key="tinj")
        self.gui.add("AIF data values", NumberListOption([0, ]), key="aif-data")
        self.gui.add("Arterial transit time (s)", NumericOption(minval=0, maxval=1.0, default=0), key="delay")
        self.gui.option("model").sig_changed.connect(self._model_changed)
        self.gui.option("aif").sig_changed.connect(self._aif_changed)
        self._aif_changed()
        self._model_changed()
        self.gui.sig_changed.connect(self._update_options)
        self._update_options()

    def _aif_changed(self):
        aif_source = self.gui.option("aif").value
        self.gui.set_visible("tinj", aif_source not in ("signal", "conc"))
        self.gui.set_visible("aif-data", aif_source in ("signal", "conc"))
        self.gui.set_visible("nt", aif_source not in ("signal", "conc"))

    def _model_changed(self):
        pass

    def _update_options(self):
        self.model.options.update(self.gui.values())



"""
Data simulation Quantiphyse plugin

Structural models, i.e. classes which return lists of different
structures and the corresponding partial volume maps

Author: Martin Craig <martin.craig@eng.ox.ac.uk>
Copyright (c) 2016-2017 University of Oxford, Martin Craig
"""

from __future__ import division, unicode_literals, absolute_import, print_function

import time

import numpy as np

from quantiphyse.data import NumpyData, DataGrid, ImageVolumeManagement
from quantiphyse.utils import QpException, get_plugins
from quantiphyse.processes import Process

from .model import Model, Parameter

def get_struc_models():
    ret = {}
    for cls in get_plugins("datasim-struc-models"):
        ret[cls.NAME] = cls
    return ret

class StructureModel(Model):
    """
    Base class for a structure model
    
    The ``structures`` property must be defined as a sequence of Parameter objects
    """

    @property
    def structures(self):
        raise NotImplementedError()

    def get_simulated_data(self, data_model, param_values, output_param_maps=False):
        """
        Generate simulated data for a given data model and parameter values

        :param data_model: DataModel which implements the ``timeseries`` method
        :param param_values: Mapping from structure name to sequence of parameter values
        :output_param_maps: If True, also output QpData containing maps of the 
                            parameter values used to generate simulated data in each voxel

        @return If ``output_param_maps`` is ``False``, QpData containing simulated timeseries data
                If ``output_param_maps`` is ``True``, tuple of simulated timeseries data and mapping
                of param name to parameter value map
        """
        raise NotImplementedError()

class PartialVolumeStructureModel(Model):
    """
    Structure model which defines the structure as a set of partial volume maps
    """

    @property
    def structure_maps(self):
        """
        :return: Mapping from name to a QpData instance containing partial volume
                 maps (range 0-1) for each known structure
        """
        raise NotImplementedError()

    def resamp(self, qpdata):
        """
        Resample a map according to resampling options
        """
        resamp_options = dict(self.options.get("resampling", {}))
        if resamp_options:
            resamp_processes = get_plugins("processes", "ResampleProcess")
            if len(resamp_processes) != 1:
                raise QpException("Can't identify Resampling process")

            ivm = ImageVolumeManagement()
            ivm.add(qpdata)
            process = resamp_processes[0](ivm)
            resamp_options.update({
                "data" : qpdata.name,
                "output-name" : "output_res"
            })
            process.execute(resamp_options)
            while process.status == Process.RUNNING:
                time.sleep(1)

            if process.status == Process.FAILED:
                raise process.exception

            # FIXME hack
            process._complete()
            return ivm.data["output_res"]
        else:
            return qpdata

    def get_simulated_data(self, data_model, param_values, output_param_maps=False):
        """
        Generic implementation to generate test data from a set of partial volume maps
        """
        if len(self.structure_maps) == 0:
            raise QpException("No structures defined")

        # First check the PV maps - they should all be in the range 0-1 and not sum to
        # more than 1 in any voxel
        sum_map = None
        for name, pv_map in self.structure_maps.items():
            pv_data = pv_map.raw()
            if not np.all(pv_data >= 0):
                raise QpException("Partial volume map contained negative values: %s" % name)
            if not np.all(pv_data <= 1):
                raise QpException("Partial volume map contained values > 1: %s" % name)
            if sum_map is None:
                sum_map = np.zeros(pv_data.shape, dtype=np.float32)
            sum_map += pv_data

        if sum_map is not None and not np.all(sum_map <= 1):
            raise QpException("Partial volume maps sum to > 1 in at least one voxel")

        output_data = None 
        for name, pv_map in self.structure_maps.items():
            struc_values = param_values[name]

            # Check that there is exactly one parameter value per structure
            single_values = {}
            for k, v in struc_values.items():
                if isinstance(v, (float, int)):
                    v = [v]
                if len(v) != 1:
                    raise QpException("This structure model cannot handle multiple parameter values in a single structure")
                single_values[k] = v[0]

            timeseries = data_model.get_timeseries(single_values, pv_map.grid.shape)
            if output_data is None:
                output_data = np.zeros(list(pv_map.grid.shape) + [timeseries.shape[-1]], dtype=np.float32)
                output_grid = pv_map.grid
                if output_param_maps:
                    param_maps = {}
                    for param in struc_values:
                        param_maps[param] = np.zeros(pv_map.grid.shape, dtype=np.float32)

            struc_data = pv_map.raw()[..., np.newaxis] * timeseries
            output_data += struc_data

            if output_param_maps:
                for param, value in single_values.items():
                    param_maps[param] += pv_map.raw() * value

        sim_data = NumpyData(output_data, grid=output_grid, name="sim_data")

        if output_param_maps:
            for param in struc_values: 
                param_maps[param] = NumpyData(param_maps[param], grid=output_grid, name=param)
            return sim_data, param_maps
        else:
            return sim_data

class UserPvModel(PartialVolumeStructureModel):
    """
    Structural model where user supplies partial volume maps

    Three default tissue types are defined: Grey matter (GM), White matter (WM) and
    CSF. A partial volume map can be specified for each of these (but they are not
    all compulsary)

    The model also supports additional user-defined structures which come in three types:

     - Additional partial volume maps. These are simply added to the existing default 
       structures. The total sum in any voxel must not exceed 1
     - Embeddings. These are added to the existing default maps, which are scaled to
       1 - the embedding partial volume. For example where the embedding partial volume
       is 1 other structures have zero partial volume. Where the embedding partial volume
       is 0.8, all other structures are scaled by multiplying by 0.2.
     - Activation mask. These are ROIs that cause a specific tissue type to replaced by
       a different set of parameter values within the ROI. For example a GM activation
       mask will replace the 'usual' GM parameters with another set of parameters within
       the mask. This mask can be used to simulate activation of a particular region, e.g.
       derived from a brain atlas.
    """
    NAME = "user"

    def __init__(self, ivm):
        PartialVolumeStructureModel.__init__(self, ivm, "User specified partial volume maps", title="Structure maps")
        self.default_strucs = [
            Parameter("gm", "Grey matter"),
            Parameter("wm", "White matter"),
            Parameter("csf", "CSF"),
        ]

    @property
    def structures(self):
        ret = [struc for struc in self.default_strucs if struc.name in self.options["pvmaps"]]
        if self.options["additional"] is not None:
            for struc in self.options["additional"].values():
                ret.append(Parameter(**struc))
        return ret

    @property
    def structure_maps(self):
        pvmaps = self.options["pvmaps"]
        try:
            ret = {}
            total_pv = None
            grid = None
            # First get the PV maps from the default structures. We also calculate the
            # sum of PVs in the maps as it may be needed when we insert embedded user-defined
            # structures. The first default structure map defines the grid FIXME should there
            # be an independent choice of grid?
            for struc in self.default_strucs:
                if struc.name in pvmaps:
                    if grid is None:
                        grid = self._ivm.data[pvmaps[struc.name]].grid
                    ret[struc.name] = self._ivm.data[pvmaps[struc.name]].resample(grid)
                    if total_pv is None:
                        total_pv = np.zeros(grid.shape, dtype=np.float32)
                    total_pv += ret[struc.name].raw()

            # Now deal with any user-defined additional structures. We handle these differently
            # depending on whether they are embeddings, activation masks or additional PV
            if self.options["additional"] is not None:
                for struc in self.options.get("additional", {}).values():
                    if grid is None:
                        grid = self._ivm.data[struc["pvmap"]].grid
                    qpdata = self._ivm.data[struc["pvmap"]].resample(grid)
                    struc_type = struc.get("struc_type", "")
                    if struc_type == "embed":
                        self._add_embedding(struc, qpdata, ret)
                    elif struc_type == "act":
                        self._add_activation_mask(struc, qpdata, ret)
                    elif struc_type == "add":
                        # Just use data directly
                        ret[struc["name"]] = qpdata
                    else:
                        raise QpException("Unknown additional structure type: %s" % struc_type)

            # Resample PV maps according to specification
            # Note that resampling/embedding/smoothing etc can lead to situations where the sum of the partial volumes
            # in the resampled maps is >1 in some voxels, even where this was not true in the
            # original maps. We need to detect this and rescale the affected voxels
            sum_maps = None
            for name in list(ret.keys()):
                resampled = self.resamp(ret[name])
                if sum_maps is None:
                    sum_maps = np.zeros(resampled.raw().shape, dtype=np.float32)
                sum_maps += resampled.raw()
                ret[name] = resampled

            if sum_maps is not None and not np.all(sum_maps <= 1):
                # Resampling has messed up the PV sum a bit - rescale to fix this but only in affected voxels
                self.debug("Max PV in resampled maps is %f: rescaling in %i voxels" % (np.max(sum_maps), np.count_nonzero(sum_maps > 1)))
                for name in list(ret.keys()):
                    pv_map = ret[name].raw()
                    pv_map[sum_maps > 1] /= sum_maps[sum_maps > 1]

            return ret
        except KeyError:
            raise

    def _add_embedding(self, struc, qpdata, strucs):
        """
        Add an embedded structure

        For an embedding, the PV map is inserted into the output unchanged and existing
        structures PVs are downweighted by the PV of the embedding (e.g. if the embedding
        has a PV of 0.5 is a voxel then existing structures in that voxel are downweighted
        by a factor of 2). This will ensure that total PV does not exceed 1

        :param struc: Structure dictionary
        :param qpdata: Embedding PV map defined on same grid as existing data
        :param strucs: Dictionary of structure name to PV map for existing structures.
                       Will be updated to include embedded structure.
        """
        pv = qpdata.raw()
        if "region" in struc:
            # If we are using only a single region of an ROI, zero out all other regions
            pv[pv != struc["region"]] = 0
        if qpdata.roi:
            # For any ROI map, voxels inside the ROI should have a PV of 1
            pv[pv > 0] = 1
        pv = pv.astype(np.float32)
        if "sigma" in struc:
            # Perform Gaussian smoothing
            pv = self._smooth_pv(NumpyData(pv, grid=qpdata.grid, name=struc["name"]), struc["sigma"])
        reweighting = 1-pv
        for name, existing_map in strucs.items():
            new_map = NumpyData(existing_map.raw() * reweighting, grid=qpdata.grid, name=name)
            strucs[name] = new_map
        strucs[struc["name"]] = NumpyData(pv, grid=qpdata.grid, name=struc["name"])

    def _smooth_pv(self, qpdata, sigma):
        """
        Do Gaussian smoothing on a partial volume map

        Typically when the map is a discrete ROI and we want to smoothly blend it into
        the other tissue PV maps

        :param qpdata: PV map
        :param sigma: Gaussian kernel std.dev in mm
        :return: Smoothed PV map as Numpy array
        """
        smooth_processes = get_plugins("processes", "SmoothingProcess")
        if len(smooth_processes) != 1:
            raise QpException("Can't identify smoothing process")

        ivm = ImageVolumeManagement()
        ivm.add(qpdata)
        process = smooth_processes[0](ivm)
        smooth_options = {
            "data" : qpdata.name,
            "sigma" : sigma,
            "output-name" : "output_smooth",
        }
        process.execute(smooth_options)
        while process.status == Process.RUNNING:
            time.sleep(1)

        if process.status == Process.FAILED:
            raise process.exception

        # FIXME hack
        process._complete()
        return ivm.data["output_smooth"].raw()

    def _add_activation_mask(self, struc, qpdata, strucs):
        """
        Add an activation mask

        This is a binary mask which splits a parent structure (e.g. GM) into two separate
        structures, inside and outside the mask, which can have different parameter properties

        :param struc: Structure dictionary. Must define the parent structure.
        :param qpdata: Activation mask map defined on same grid as existing data
        :param strucs: Dictionary of structure name to PV map for existing structures.
                       Will be updated to include split parent structure.
        """
        # Activation mask - replace parent structure
        parent_struc = struc.get("parent_struc", None)
        if parent_struc is None:
            raise QpException("Parent structure not defined for activation mask: %s" % struc["name"])
        elif parent_struc not in strucs:
            raise QpException("Parent structure '%s' not found in structures list for activation mask: %s" % (parent_struc, struc["name"]))
        activation_mask = np.copy(qpdata.raw())
        if "region" in struc:
            # If a specific region is given, isolate it
            activation_mask[activation_mask != struc["region"]] = 0

        if qpdata.roi:
            # If mask is an ROI, make it take values 0 and 1
            activation_mask[activation_mask <= 0] = 1
            activation_mask[activation_mask > 0] = 1

        # Activation structure takes over parent structure within the mask.
        # We do this by multiplication so the activation mask can in principle
        # be probabilistic
        activation_mask = activation_mask.astype(np.float32)
        parent_qpdata = strucs[parent_struc]
        parent_data = np.copy(parent_qpdata.raw())
        activation_data = np.copy(parent_qpdata.raw())
        activation_data *= activation_mask
        parent_data *= (1-activation_mask)
        strucs[parent_struc] = NumpyData(parent_data, grid=parent_qpdata.grid, name=parent_qpdata.name)
        strucs[struc["name"]] = NumpyData(activation_data, grid=parent_qpdata.grid, name=struc["name"])

class FastStructureModel(PartialVolumeStructureModel):
    """
    Structural model which derives partial volume maps from a FAST segmentation
    """
    NAME = "fast"

    def __init__(self, ivm):
        StructureModel.__init__(self, ivm, "Partial volume maps from a FAST segmentation")
       
    @property
    def structures(self):
        return [
            Parameter("gm", "Grey matter"),
            Parameter("wm", "White matter"),
            Parameter("csf", "CSF"),
        ]

    @property
    def structure_maps(self):
        processes = get_plugins("processes", "FastProcess")
        if len(processes) != 1:
            raise QpException("Can't identify Fast process")
        
        struc = self.options.get("struc", None)
        if struc not in self._ivm.data:
            raise QpException("Structural image not loaded: %s" % struc)
        
        qpdata = self._ivm.data[struc]
        ivm = ImageVolumeManagement()
        ivm.add(qpdata)
        process = processes[0](ivm)
        fast_options = {
            "data" : qpdata.name,
            "class" : 3,
            "type" : self.options["type"],
            "output-pve" : True,
            "output-pveseg" : False,
        }
        process.execute(fast_options)
        while process.status == Process.RUNNING:
            time.sleep(1)

        if process.status == Process.FAILED:
            raise process.exception

        # FIXME hack
        process._complete()

        return {
            "gm" : ivm.data["%s_pve_1" % qpdata.name],
            "wm" : ivm.data["%s_pve_2" % qpdata.name],
            "csf" : ivm.data["%s_pve_0" % qpdata.name],
        }

def fslimage_to_qpdata(img, name=None, vol=None, region=None):
    """ Convert fsl.data.Image to QpData """
    if not name: name = img.name
    if vol is not None:
        data = img.data[..., vol]
    else:
        data = img.data
    if region is not None:
        data = (data == region).astype(np.int)
    return NumpyData(data, grid=DataGrid(img.shape[:3], img.voxToWorldMat), name=name)

class FslStdStructureModel(PartialVolumeStructureModel):
    """
    Structural model using standard FSL structural data

    FIXME not functional at present - not clear that FSL supplies relevant
    segmentation data out of the box!
    """
    NAME = "fsl"

    ATLAS_PREFIXES = [
        "MNI",
    ]

    def __init__(self, ivm):
        StructureModel.__init__(self, ivm, "FSL MNI standard data")
        from fsl.data.atlases import AtlasRegistry
        self._registry = AtlasRegistry()
        self._registry.rescanAtlases()
        atlas_names = []
        self._atlases = {}
        for atlas in sorted(self._registry.listAtlases(), key=lambda x: x.name):
            for prefix in self.ATLAS_PREFIXES:
                if atlas.name.startswith(prefix):
                    for pixdim in atlas.pixdims:
                        name = atlas.name + " %.2gx%.2gx%.2g mm" % pixdim
                        self._atlases[name] = (atlas, pixdim)

    @property
    def structures(self):
        atlas, pixdims = self._atlases[self.options["atlas"]]
        structures = []
        for label in atlas.labels:
            structures.append(Parameter(label.name, label.name))
        return structures

    @property
    def structure_maps(self):
        atlas, pixdims = self._atlases[self.options["atlas"]]
        structure_maps = {}
        atlas_map = self._registry.loadAtlas(atlas.atlasID, loadSummary=False, resolution=pixdims[0])
        for idx, label in enumerate(atlas.labels):
            structure_maps[label.name] = fslimage_to_qpdata(atlas_map, vol=idx, name=label.name)
        return structure_maps

class CheckerboardModel(StructureModel):
    """
    Model which builds a checkerboard for up to 3 varying parameters

    This model differs from the usual 'partial volume' type models in that
    it expects multiple values for up to 3 parameters. It builds a 1/2/3D
    grid in which each varying parameter changes along a given dimension.
    There can only be a single 'structure' defined whose name is arbitrary.
    """
    NAME = "checkerboard"

    def __init__(self, ivm):
        StructureModel.__init__(self, ivm, "Checkerboard")

    @property
    def structures(self):
        return [
            Parameter("data", "Sequence of test values"),
        ]

    def get_simulated_data(self, data_model, param_values, output_param_maps=False):
        if len(param_values) != 1:
            raise QpException("Can only have a single structure in the checkerboard model")
        param_values = list(param_values.values())[0]

        param_values_list = {}
        varying_params = []
        for param, values in param_values.items():
            if isinstance(values, (int, float)):
                values = [values]
            if len(values) > 1:
                varying_params.append(param)
            param_values_list[param] = values

        num_varying_params = len(varying_params)
        if num_varying_params > 3:
            raise QpException("At most 3 parameters can vary")
        elif num_varying_params == 0:
            # Make it a square for simplicity
            num_varying_params = 2

        voxels_per_patch = self.options.get("voxels-per-patch", 100)
        side_length = int(round(voxels_per_patch ** (1.0 / float(num_varying_params))))
        patch_dims = [side_length] * num_varying_params
        while len(patch_dims) < 3:
            patch_dims += [1,]

        repeats = [[0], [0], [0]]
        checkerboard_dims = []
        for idx, param in enumerate(varying_params):
            num_values = len(param_values_list[param])
            repeats[idx] = range(num_values)
            checkerboard_dims.append(patch_dims[idx] * num_values)
        for idx in range(len(varying_params), 3):
            checkerboard_dims.append(patch_dims[idx])

        output_data = None
        import itertools
        for indexes in itertools.product(*repeats):
            patch_values = dict(param_values)
            for idx, param in enumerate(varying_params):
                patch_values[param] = patch_values[param][indexes[idx]]
            
            timeseries = data_model.get_timeseries(patch_values)
            if output_data is None:
                output_data = np.zeros(list(checkerboard_dims) + [len(timeseries)], dtype=np.float32)
                if output_param_maps:
                    param_maps = {}
                    for param in param_values:
                        param_maps[param] = np.zeros(checkerboard_dims, dtype=np.float32)

            slices = []
            for dim_idx, patch_idx in enumerate(indexes):
                dim_length = patch_dims[dim_idx]
                slices.append(slice(patch_idx*dim_length, (patch_idx+1)*dim_length))
            output_data[slices] = timeseries
            if output_param_maps:
                for param, value in patch_values.items():
                    param_maps[param][slices] = value

        grid = DataGrid(checkerboard_dims, np.identity(4))
        sim_data = NumpyData(output_data, grid=grid, name="sim_data")

        if output_param_maps:
            for param in param_values: 
                param_maps[param] = NumpyData(param_maps[param], grid=grid, name=param)
            return sim_data, param_maps
        else:
            return sim_data



"""
Data simulation Quantiphyse plugin

Data models, i.e. classes which generated simulated data for different
types of imaging

Author: Martin Craig <martin.craig@eng.ox.ac.uk>
Copyright (c) 2016-2017 University of Oxford, Martin Craig
"""

from __future__ import division, unicode_literals, absolute_import, print_function

import logging

import numpy as np

from fabber import Fabber

from quantiphyse.data import NumpyData
from quantiphyse.utils import get_plugins, QpException

from .model import Model, Parameter

LOG = logging.getLogger(__name__)

def get_data_models():
    ret = {}
    for cls in get_plugins("datasim-data-models"):
        ret[cls.NAME] = cls
    return ret

class DataModel(Model):
    """
    Implements a model for producing simulated data

    A data model has the ability to generate timeseries data from a dictionary
    of parameter values.
    """

    def __init__(self, ivm, title):
        Model.__init__(self, ivm, title)

        # Include some basic parameters common to many models
        self.known_params = [
            Parameter(
                "t1",
                "Tissue T1",
                default=1.3,
                units="s",
                struc_defaults={
                    "gm" : 1.3,
                    "wm" : 1.1,
                    "csf" : 4.3,
                },
                aliases=["T_1"],
            ),
            Parameter(
                "t2",
                "Tissue T2",
                default=100,
                units="ms",
                struc_defaults={
                    "gm" : 100,
                    "wm" : 50,
                    "csf" : 750,
                },
                aliases=["T_2"],
            ),
            Parameter(
                "pc",
                "Partition coefficient",
                default=0.9,
                units="",
                struc_defaults={
                    "gm" : 0.98,
                    "wm" : 0.82,
                    "csf" : 1.15,
                },
                aliases=["lambda"],
            ),
            Parameter(
                "t1b",
                "Blood T1",
                default=1.65,
                units="s",
                struc_defaults={
                    "gm" : 1.65,
                    "wm" : 1.65,
                    "csf" : 1.65,
                },
                aliases=["T1_b", "T_1b"],
            ),
        ]

    def get_timeseries(self, param_values, shape=None):
        """
        Get the data model time series for a set of parameter values

        :param param_values: Mapping from parameter name to value
        :param shape: If specified, 3D output shape for the data being constructed.
                      Data models may return either a 1D timeseries to 
                      be used for all voxels, or a 4D voxelwise map. If shape is None
                      a 1D timeseries must be returned
        :return: 1D or 4D np.array containing the timeseries
        """
        raise NotImplementedError()

class SpinEchoDataModel(Model):
    """
    Implements a model for producing simulated spin-echo data
    """
    NAME = "spin_echo"

    def __init__(self, ivm):
        DataModel.__init__(self, ivm, "Spin Echo")

    def get_timeseries(self, param_values, shape=None):
        tr = self.options["tr"]
        te = self.options["te"]
        m0 = self.options.get("m0", 1000)
        t1 = param_values.get("t1", 1.3)
        t2 = param_values.get("t2", 100)
        pc = param_values.get("pc", 1)
        # FIXME proton density?
        return np.array([m0 * pc * (1-np.exp(-tr/t1)) * np.exp(-te/t2)])

    @property
    def params(self):
        return [param for param in self.known_params if param.name in ("t1", "t2", "pc")]

class FabberDataModel(DataModel):
    """
    Generates simulated data using Fabber
    """
    NAME = ""

    def __init__(self, ivm, title):
        DataModel.__init__(self, ivm, title)
        from fabber import Fabber
        search_dirs = get_plugins(key="fabber-dirs")
        self._fab = Fabber(*search_dirs)

    @property
    def params(self):
        return [param for param in self.known_params if param.name in self.real_param_names]

    @property
    def real_param_names(self):
        """
        Fabber models often use different names for the same basic parameter (e.g. T1 might be
        named as t1, T1, T_1, ...). This is obviously a pain but we get around it by allowing
        parameters to have 'aliases' and maintaining a mapping from the 'standard' name to the one
        used in the model
        """
        model_params = self._fab.get_model_params(self.fab_options)
        real_param_names = {}
        for param in model_params:
            # Note that we do not worry if Fabber is returning parameters that are not in our list of
            # known parameters. The data model is responsible for adding in values for these and if it
            # does not an error will occur on evaluation of the model. For example static tissue signal
            # in ASL may be derived from tissue properties (T1, T2 etc)
            for known_param in self.known_params:
                if param.lower() == known_param.name.lower() or param in known_param.kwargs.get("aliases", []):
                    real_param_names[known_param.name] = param
                    break
        return real_param_names

    def get_timeseries(self, param_values, shape=None):
        LOG.debug("Fabbber options %s", self.fab_options)
        real_param_values = {}
        real_param_names = self.real_param_names
        # Need to correct parameter names from the 'standard' names to those
        # used in the model
        for k, v in param_values.items():
            real_param_values[real_param_names[k]] = v
        ts = self._fab.model_evaluate(self.fab_options, real_param_values, self.nt)
        LOG.debug("Fabbber timeseries %s", ts)
        return np.array(ts)

    @property
    def fab_options(self):
        raise NotImplementedError()

    @property
    def nt(self):
        raise NotImplementedError()

class AslDataModel(FabberDataModel):
    """
    Generates simulated ASL data using Fabber

    This uses the resting-state ASL model 'aslrest'
    """
    NAME = "asl"
    
    def __init__(self, ivm):
        FabberDataModel.__init__(self, ivm, "Arterial Spin Labelling")
        self.slicet = 0
        self.known_params += [
            Parameter(
                "ftiss", 
                "CBF",
                default=10.0,
                units="ml/100g/s",
                struc_defaults={
                    "gm" : 50.0,
                    "wm" : 10.0,
                    "csf" : 0.0,
                }
            ),
            Parameter(
                "delttiss",
                "ATT to tissue",
                default=1.3,
                units="s",
                struc_defaults={
                    "gm" : 1.3,
                    "wm" : 1.6,
                    "csf" : 1.3,
                },
            ),
            Parameter(
                "fblood",
                "Blood CBF",
                default=10.0,
                units="ml/100g/s",
                struc_defaults={}
            ),
            Parameter(
                "deltblood",
                "Transit time to artery",
                default=1.0,
                units="s",
                struc_defaults={}
            ),
        ]

    def get_timeseries(self, param_values, shape=None):
        if shape is not None and self.options["slicedt"] != 0:
            ret = np.zeros(list(shape) + [self.nt], dtype=np.float32)
            for z in range(shape[2]):
                self.slicet = z*float(self.options["slicedt"])/1000
                ret[:, :, z, :] = self._get_slice_timeseries(param_values)
            self.slicet = 0
        else:
            self.slicet = 0
            ret = self._get_slice_timeseries(param_values)
        return ret

    def _get_slice_timeseries(self, param_values):
        ts = FabberDataModel.get_timeseries(self, param_values)
        ts = self.options["m0"] * self.options["alpha"] * ts / 6000
        nrpt = self.options["repeats"]

        # For tag control pairs calculate static tissue signal at each
        # PLD and add to Fabber timeseries. Note that fabber by default
        # uses zero for control images and by default groups by repeats
        if self.options["iaf"] == "tc":
            tis = np.array(self.options["plds"]) + self.options["tau"] + self.slicet
            tr = self.options["tr"]
            te = self.options["te"]
            t1 = param_values.get("t1", 1.3)
            t2 = param_values.get("t2", 100)
            t1b = param_values.get("t1b", 1.65)
            stattiss = np.tile(self.options["pct"] * self.options["m0"] * (1-np.exp(-tr/t1)) * np.exp(-te/t2), len(tis))
            tc = []
            for idx, sig in enumerate(ts):
                tc.append(stattiss[int(idx/(2*nrpt))] - sig)
            ts = np.array(tc)

        ret = np.tile(ts, nrpt)

        # If we are grouping by TIs we need to reorder the timeseries so all the
        # full set of TIs is together and then repeated
        if self.options["ibf"] == "tis":
            reordered = []
            npld = len(self.options["plds"])
            ntc = 1 if self.options["iaf"] == "diff" else 2
            for pld in range(npld):
                for rpt in range(nrpt):
                    for tc in range(ntc):
                        reordered.append(ret[pld*nrpt*ntc + rpt*ntc + tc])
            ret = np.array(reordered)

        return ret

    @property
    def fab_options(self):
        fab_options = {
            "model" : "aslrest",
            "inctiss" : True,
            "incbat" : True,
            "inct1" : True,
        }
        plds = self.options["plds"]
        for idx, pld in enumerate(plds):
            fab_options["pld%i" % (idx+1)] = pld + self.slicet
        fab_options.update(self.options)

        for opt in ("m0", "alpha", "pct", "slicedt"):
            fab_options.pop(opt, None)

        return fab_options

    @property
    def nt(self):
        diff = self.options["iaf"] == "diff"
        return len(self.options["plds"]) * (1 if diff else 2) * self.options["repeats"]
           
class DscDataModel(FabberDataModel):
    """
    Generates simulated DSC data
    """
    NAME = "dsc"
    
    def __init__(self, ivm):
        FabberDataModel.__init__(self, ivm, "Dynamic Susceptibility Contrast")

        self.known_params = [
            Parameter("sig0", "Signal offset", default=100.0),
            Parameter("cbf", "CBF", default=10.0),
        ]

    @property
    def fab_options(self):
        fab_options = {
            "model" : "dsc",
        }
        fab_options.update(self.options)
        return fab_options

    @property
    def nt(self):
        return len(self.options["aif"])
         
class DceDataModel(FabberDataModel):
    """
    Generates simulated DCE data
    """
    NAME = "dce"
    
    def __init__(self, ivm):
        FabberDataModel.__init__(self, ivm, "Dynamic Contrast-Enhanced MRI")
        
        from fabber import Fabber
        search_dirs = get_plugins(key="fabber-dirs")
        self._fab = Fabber(*search_dirs)

        self.known_params = [
            Parameter("sig0", "Signal offset", default=100.0),
            Parameter("fp", "Flow, Fp", default=0.5, units="min^-1"),
            Parameter("ps", "Permeability-surface area, PS", default=0.5, units="min^-1"),
            Parameter("ktrans", "Transfer coefficient, Ktrans", default=0.5, units="min^-1"),
            Parameter("ve", "Extracellular volume fraction Ve", default=0.2),
            Parameter("vp", "Vascular plasma volume fraction Vp", default=0.05),
            Parameter("t10", "T1", default=1.0, units="s"),
        ]

    @property
    def fab_options(self):
        fab_options = {
            "infer-sig0" : True,
            "infer-fp" : True,
            "infer-ps" : True,
            "infer-t10" : True,
            "infer-delay" : False,
        }
        fab_options.update(self.options)
        if fab_options["model"] == "dce_ETM":
            fab_options["model"] = "dce_tofts"
            fab_options["infer-vp"] = True

        # Transit delay time to include injection time for population AIF
        if "tinj" in fab_options:
            fab_options["delay"] = fab_options["delay"] + fab_options.pop("tinj")

        # Times in minutes and TR in s
        fab_options["delt"] = fab_options["delt"] / 60
        fab_options["delay"] = fab_options["delay"] / 60
        fab_options["tr"] = fab_options["tr"] / 1000

        # NT is not a Fabber option, it is handled separately
        fab_options.pop("nt", None)
        if "aif-data" not in fab_options:
            fab_options["aif-data"] = [1.0]

        return fab_options

    @property
    def nt(self):
        return self.options.get("nt", len(self.options.get("aif-data", [1])))
         
         


"""
Quantiphyse plugin for data simulation

Author: Martin Craig <martin.craig@eng.ox.ac.uk>
Copyright (c) 2016-2017 University of Oxford, Martin Craig
"""
from .widget import DataSimWidget
from .process import DataSimProcess
from .data_models import *
from .data_model_views import *
from .struc_models import *
from .struc_model_views import *

def get_view_class(model_class):
    view_name = model_class.__name__ + "View"
    return globals().get(view_name, None)

QP_MANIFEST = {
    "widgets" : [
        DataSimWidget
    ],
    "processes" : [
        DataSimProcess
    ],
    "datasim-data-models" : [
        SpinEchoDataModel,
        AslDataModel,
        DscDataModel,
        DceDataModel,
    ],
    "datasim-struc-models" : [
        UserPvModel,
        CheckerboardModel,
        #FslStdStructureModel,
        FastStructureModel,
    ]
}
